#!/usr/bin/env python
"""
convert output data of tomte's in java implemented learner to a concreet uppaal model

learned data :
   json file :  specifies learned abstractions and input,output alphabet sut
    dot file :  specifies learned abstract model 
"""


# Author: Harco Kuppens


import sys, re, pprint  # modules from standard lib (python 2.6 and later)
import argparse

import pp_xml  # local module (which only uses standard lib modules)
import efsm    # local module ; next to standard lib modules it uses: pyuppaal module -> locally installed
import util
import efsm_state
import efsm_statevar

def learnresult2uppaal_firstlast(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead):
    """
     convert output data of tomte's in java implemented learner to a concreet uppaal model

     learned data :
        json file :  specifies learned abstractions and input,output alphabet sut
        dot file :   specifies learned abstract model
    """



    [start_state,transitions]=efsm.get_lts_from_dotfile(dot_filename_in)

    #TODO: get below info also from yaml file!!
    [inputs,outputs,cons,interfaces]=efsm.read_abstractions_from_jsonfile(json_filename_in)
    # min max value in json are used for learning

    number_params={}
    for i in interfaces:
       number_params[ i['method_name'] ] = len(i['param_types'])

    transitions=efsm.parse_labels_of_lts(transitions)
    transitions=efsm.apply_mapper_data_on_transitions(transitions,inputs,outputs,cons,number_params)
    transitions=efsm.split_input_and_output(transitions)

    model=efsm.get_uppaal_model(start_state,transitions,interfaces,cons)

    f=open(model_file_out,'w')
    f.write( pp_xml.pformat(model.to_xml(),step="    ") )
    f.close()

def learnresult2uppaal_state(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead):
    """
     convert output data of tomte's in java implemented learner to a concreet uppaal model

     learned data :
        json file :  specifies learned abstractions and input,output alphabet sut
        dot file :   specifies learned abstract model 
    """

    efsm=efsm_state

    [start_state,transitions]=efsm.get_lts_from_dotfile(dot_filename_in)

    #TODO: get below info also from yaml file!!
    [input_abstractions,output_abstractions,cons,interfaces]=efsm.read_abstractions_from_jsonfile(json_filename_in)
    # min max value in json are used for learning

    number_params={}
    for i in interfaces:
        number_params[ i['method_name'] ] = len(i['param_types'])

    transitions=efsm.parse_labels_of_lts(transitions)

    
    #pprint.pprint(transitions)
    #pprint.pprint(input_abstractions)
    #pprint.pprint(output_abstractions)
    #sys.exit()

    (transitions,statevars,abstractinputs)=efsm.apply_mapper_data_on_transitions(transitions,input_abstractions,output_abstractions,cons,number_params,giveAbstractInputsInstead)
    transitions=efsm.split_input_and_output(transitions)

    model=efsm.get_uppaal_model(start_state,transitions,interfaces,statevars,giveAbstractInputsInstead,abstractinputs,cons)

    f=open(model_file_out,'w')
    f.write( pp_xml.pformat(model.to_xml(),step="    ") )
    f.close()

def learnresult2uppaal_statevar(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead):
    """
     convert output data of tomte's in java implemented learner to a concreet uppaal model

     learned data :
        json file :  specifies learned abstractions and input,output alphabet sut
        dot file :   specifies learned abstract model 
    """

    efsm=efsm_statevar

    [start_state,transitions]=efsm.get_lts_from_dotfile(dot_filename_in)

    #TODO: get below info also from yaml file!!
    [input_abstractions,output_abstractions,cons,interfaces]=efsm.read_abstractions_from_jsonfile(json_filename_in)
    # min max value in json are used for learning

    number_params={}
    for i in interfaces:
        number_params[ i['method_name'] ] = len(i['param_types'])

    transitions=efsm.parse_labels_of_lts(transitions)

    #pprint.pprint(transitions)
    #pprint.pprint(input_abstractions)
    #pprint.pprint(output_abstractions)
    #sys.exit()

    (transitions,statevars,abstractinputs)=efsm.apply_mapper_data_on_transitions(transitions,input_abstractions,output_abstractions,cons,number_params,giveAbstractInputsInstead)
    transitions=efsm.split_input_and_output(transitions)

    model=efsm.get_uppaal_model(start_state,transitions,interfaces,statevars,giveAbstractInputsInstead,abstractinputs,cons)

    f=open(model_file_out,'w')
    f.write( pp_xml.pformat(model.to_xml(),step="    ") )
    f.close()

def args_parsing(args):
    import os.path
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('jsonfile', help='json file containing learned abstractions')
    argparser.add_argument('dotfile', help='graphviz dot file containing learned abstract model')

    argparser.add_argument('modelfile', help='output efsm in uppaal model file')
    argparser.add_argument('mode', help='mode of learning, eg firstlast mode or state mode')

    argparser.add_argument('output', help='type of output: abstractInputs or concreteOutputs ')



    parse_result=argparser.parse_args(args[1:])


    # handle input/output file argument
    json_filename_in=os.path.abspath(parse_result.jsonfile).replace('\\','/')
    dot_filename_in=os.path.abspath(parse_result.dotfile).replace('\\','/')
    model_file_out=os.path.abspath(parse_result.modelfile).replace('\\','/')
    # Note: on windows also interprets / as dir sep, and returns path with windows \
    mode=parse_result.mode
    output=parse_result.output

    util.check_isreadablefile(json_filename_in)
    util.check_isreadablefile(dot_filename_in)

    return (json_filename_in,dot_filename_in,model_file_out,mode,output)

def main(args):
    """
     convert output data of tomte's in java implemented learner to a concreet uppaal model

     learned data :
        json file :  specifies learned abstractions and input,output alphabet sut
        dot file :   specifies learned abstract model 
    """
    (json_filename_in,dot_filename_in,model_file_out,mode,output)=args_parsing(args)
    if mode=="state":
      giveAbstractInputsInstead=False
      if output=="abstractInputs":
         giveAbstractInputsInstead=True
      learnresult2uppaal_state(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead)
    elif mode=="stateVar":
      giveAbstractInputsInstead=False
      if output=="abstractInputs":
         giveAbstractInputsInstead=True
      learnresult2uppaal_statevar(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead)
    else:
      learnresult2uppaal_firstlast(json_filename_in,dot_filename_in,model_file_out,giveAbstractInputsInstead)




if __name__ == "__main__":
    main(sys.argv)
